(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2011 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)


module Make(V:Constant.S) =
  struct
    module A = ARMArch.Make(V)
    open A
    open A.Out
    open Printf

(* No addresses in code *)
    let extract_addrs _ins = StringSet.empty

(************************)
(* Template compilation *)
(************************)
    type cond = EQ | NE 

    let pp_cond = function
      | EQ -> "eq"
      | NE -> "ne"

    let pp_cond_opt = function
      | None -> ""
      | Some c -> pp_cond c

    let is_cond = function
      | None -> false
      | Some _ -> true

    let pp_setflags = function
      | SetFlags -> "s"
      | DontSetFlags -> ""

(* Arithmetic *)
    let op3regs memo s c rD rA rB =
      let memo =
        sprintf "%s%s%s" 
          memo (pp_setflags s) (pp_cond_opt c) in
      { empty_ins with
        memo=memo^ " ^o0,^i0,^i1";
        inputs=[rA; rB];
        outputs=[rD]; cond=is_cond c; }

    let op2regsI memo s c rD rA i =
      let memo =
        sprintf "%s%s%s"
          memo (pp_setflags s) (pp_cond_opt c) in
      { empty_ins with
        memo= sprintf "%s ^o0,^i0,#%i" memo i;
        inputs=[rA];
        outputs=[rD]; cond=is_cond c; }

(* Moves *)
    let movi c r1 i =
      let memo = sprintf "%s%s" "mov" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^o0,#%i" memo i ;
        inputs = [] ;
        outputs = [r1]; cond=is_cond c; }

    let mov c r1 r2 =
      let memo = sprintf "%s%s" "mov" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^o0,^i0" memo ;
        inputs = [r2] ;
        outputs = [r1]; cond=is_cond c; }


(* Memory *)
    let ldr2 c r1 r2 =
      let memo = sprintf "%s%s" "ldr" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^o0,[^i0]" memo ;
        inputs = [r2] ;
        outputs = [r1] ; cond=is_cond c; }

    let ldr3 c r1 r2 r3 =
      let memo = sprintf "%s%s" "ldr" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^o0,[^i0,^i1]" memo ;
        inputs = [r2;r3] ;
        outputs = [r1] ; cond=is_cond c; }

    let str2 c r1 r2 =
      let memo = sprintf "%s%s" "str" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^i0,[^i1]" memo ;
        inputs = [r1;r2] ;
        outputs = [] ; cond=is_cond c; }

    let str3 c r1 r2 r3 =
      let memo = sprintf "%s%s" "str" (pp_cond_opt c) in
      { empty_ins with
        memo = sprintf "%s ^i0,[^i1,^i2]" memo ;
        inputs = [r1;r2;r3] ;
        outputs = [] ; cond=is_cond c; }

(* cmp & branches *)
    let cmpi r1 k =
      { empty_ins with
        memo = sprintf "cmp ^i0,#%i" k ;
        inputs = [r1;] ;
        outputs = [] ;
      }

    let cmp r1 r2 =
      { empty_ins with
        memo = "cmp ^i0,^i1" ;
        inputs = [r1;r2;] ;
        outputs = [] ;
      }

    let b lbl =
      { empty_ins with memo = sprintf "b %s" (A.Out.dump_label lbl) ;
        branch=[Branch lbl] ; }

    let bcc cond lbl =
      { empty_ins with
        memo = sprintf "b%s %s" (pp_cond cond) (A.Out.dump_label lbl) ;
        branch=[Next; Branch lbl] ; }

    let emit_lbl lbl =
      { empty_ins with
        memo=sprintf "%s:" (A.Out.dump_label lbl) ;
        label = Some lbl ; branch=[Next] ; }

    let decr r i =
      { empty_ins with
        memo = sprintf "subs ^o0,^i0,#%i" i ;
        inputs = [r;] ; outputs = [r;] ; }

    let emit_loop k =
      let lbl1 = Label.next_label "L" in
      let lbl2 = Label.next_label "L" in
      cmpi loop_idx 0::
      b lbl2::
      emit_lbl lbl1::
      k@
      [ decr loop_idx 1;
        emit_lbl lbl2 ;
        bcc NE lbl1; ]

    let compile_ins ins k = match ins with
(* Arithmetic *)
    | I_ADD (s,r1, r2, i) ->  op2regsI "add" s None r1 r2 i::k
    | I_SUB (s,r1, r2, i) ->  op2regsI "sub" s None r1 r2 i::k
    | I_AND (s,r1, r2, i) ->  op2regsI "and" s None r1 r2 i::k
    | I_ADD3 (s,r1, r2, r3) ->  op3regs "add" s None r1 r2 r3::k
    | I_SUB3 (s,r1, r2, r3) ->  op3regs "sub" s None r1 r2 r3::k
    | I_XOR (s,r1, r2, r3) -> op3regs "eor" s None r1 r2 r3::k
(* Moves *)
    | I_MOVI (r, i) -> movi None r i::k
    | I_MOVINE (r, i) ->  movi (Some NE) r i::k
    | I_MOVIEQ (r,i) -> movi (Some EQ) r i::k
    | I_MOV (r1,r2) -> mov None r1 r2::k
    | I_MOVNE (r1, r2) ->  mov (Some NE) r1 r2::k
    | I_MOVEQ (r1, r2) -> mov (Some EQ) r1 r2::k
(* Memory *)
    | I_LDR (r1, r2) ->  ldr2 None r1 r2::k
    | I_LDRNE (r1, r2) ->  ldr2 (Some NE) r1 r2::k
    | I_LDR3 (r1, r2, r3) ->  ldr3 None r1 r2 r3::k
    | I_STR (r1, r2) ->  str2 None r1 r2::k
    | I_STRNE (r1, r2) ->  str2 (Some NE) r1 r2::k
    | I_STR3 (r1, r2, r3) ->  str3 None r1 r2 r3::k
(* Comparisons and branches *)
    | I_CMPI (r1, i) -> cmpi r1 i::k
    | I_CMP (r1, r2)-> cmp r1 r2::k
    | I_B lbl -> b lbl::k
    | I_BNE lbl -> bcc NE lbl::k
    | I_BEQ lbl -> bcc EQ lbl::k
(* Misc *)
    | I_DMB ->
        { empty_ins with memo = "dmb"; }::k
  end
